from typing import Any, Tuple

import httpx
from pydantic import AnyUrl

from enlyze.api_clients.base import ApiBaseClient, PaginatedResponseBaseModel
from enlyze.constants import TIMESERIES_API_SUB_PATH


class _PaginatedResponse(PaginatedResponseBaseModel):
    next: AnyUrl | None
    data: list[Any] | dict[str, Any]


class TimeseriesApiClient(ApiBaseClient[_PaginatedResponse]):
    """Client class encapsulating all interaction with the Timeseries API

    :param token: API token for the ENLYZE platform
    :param base_url: Base URL of the ENLYZE platform
    :param timeout: Global timeout for all HTTP requests sent to the Timeseries API

    """

    PaginatedResponseModel = _PaginatedResponse

    def __init__(
        self,
        *,
        token: str,
        base_url: str | httpx.URL,
        **kwargs: Any,
    ):
        super().__init__(
            token=token,
            base_url=httpx.URL(base_url).join(TIMESERIES_API_SUB_PATH),
            **kwargs,
        )

    def _transform_paginated_response_data(
        self, paginated_response_data: list[Any] | dict[str, Any]
    ) -> list[dict[str, Any]]:
        # The timeseries endpoint's response data field is a mapping.
        # Because get_paginated assumes the ``data`` field to be a list,
        # we wrap it into a list.
        return (
            paginated_response_data
            if isinstance(paginated_response_data, list)
            else [paginated_response_data]
        )

    def _has_more(self, paginated_response: _PaginatedResponse) -> bool:
        return paginated_response.next is not None

    def _next_page_call_args(
        self,
        *,
        url: str,
        params: dict[str, Any],
        paginated_response: _PaginatedResponse,
        **kwargs: Any,
    ) -> Tuple[str, dict[str, Any], dict[str, Any]]:
        next_url = str(paginated_response.next)
        return (next_url, params, kwargs)



import logging
from datetime import datetime, timezone
from typing import Sequence

import enlyze.models as user_models
from enlyze.constants import MINIMUM_RESAMPLING_INTERVAL
from enlyze.errors import EnlyzeError, ResamplingValidationError

VARIABLE_ARRAY_DATA_TYPES = (
    user_models.VariableDataType.ARRAY_BOOLEAN,
    user_models.VariableDataType.ARRAY_STRING,
    user_models.VariableDataType.ARRAY_INTEGER,
    user_models.VariableDataType.ARRAY_FLOAT,
)

_NAIVE_DATETIME_DISCOURAGED_LOG_MESSAGE = (
    "Passing naive datetime is discouraged, assuming local timezone."
)


def _ensure_datetime_aware(dt: datetime) -> datetime:
    """Make the returned datetime timezone aware.

    Naive datetime will be assumed to be in local timezone and then converted to aware
    datetime expressed in UTC.

    """
    return dt.astimezone(timezone.utc)


def validate_datetime(dt: datetime) -> datetime:
    if dt.utcoffset() is None:
        logging.warning(_NAIVE_DATETIME_DISCOURAGED_LOG_MESSAGE)

    return _ensure_datetime_aware(dt)


def validate_start_and_end(start: datetime, end: datetime) -> tuple[datetime, datetime]:
    if start.utcoffset() is None or end.utcoffset() is None:
        logging.warning(_NAIVE_DATETIME_DISCOURAGED_LOG_MESSAGE)

    start = _ensure_datetime_aware(start)
    end = _ensure_datetime_aware(end)
    if start >= end:
        raise EnlyzeError("Start must be earlier than end.")
    return start, end


def validate_timeseries_arguments(
    start: datetime,
    end: datetime,
    variables: Sequence[user_models.Variable],
) -> tuple[datetime, datetime, str]:
    if not variables:
        raise EnlyzeError("Need to request at least one variable")

    start, end = validate_start_and_end(start, end)

    machine_uuids = frozenset(v.machine.uuid for v in variables)

    if len(machine_uuids) != 1:
        raise EnlyzeError(
            "Cannot request timeseries data for more than one machine per request."
        )

    return start, end, str(next(iter(machine_uuids)))


def validate_resampling_interval(
    resampling_interval: int,
) -> None:
    if resampling_interval < MINIMUM_RESAMPLING_INTERVAL:
        raise ResamplingValidationError(
            "resampling_interval must be greater than or equal to"
            f" {MINIMUM_RESAMPLING_INTERVAL}, {resampling_interval=} provided."
        )


def validate_resampling_method_for_data_type(
    resampling_method: user_models.ResamplingMethod,
    data_type: user_models.VariableDataType,
) -> None:
    if data_type in VARIABLE_ARRAY_DATA_TYPES:
        raise ResamplingValidationError(
            f"Cannot resample {data_type=} as it is an array variable data type"
        )

    if data_type in (
        user_models.VariableDataType.BOOLEAN,
        user_models.VariableDataType.STRING,
    ) and resampling_method in (
        user_models.ResamplingMethod.SUM,
        user_models.ResamplingMethod.AVG,
        user_models.ResamplingMethod.MEDIAN,
    ):
        raise ResamplingValidationError(
            f"{data_type=} cannot be resampled with {resampling_method=}"
        )



from typing import Any

import httpx
from pydantic import BaseModel

from enlyze.api_clients.base import ApiBaseClient, PaginatedResponseBaseModel
from enlyze.constants import PRODUCTION_RUNS_API_SUB_PATH


class _Metadata(BaseModel):
    next_cursor: int | None
    has_more: bool


class _PaginatedResponse(PaginatedResponseBaseModel):
    metadata: _Metadata
    data: list[dict[str, Any]]


class ProductionRunsApiClient(ApiBaseClient[_PaginatedResponse]):
    """Client class encapsulating all interaction with the Production Runs API

    :param token: API token for the ENLYZE platform
    :param base_url: Base URL of the ENLYZE platform
    :param timeout: Global timeout for all HTTP requests sent to the Production Runs API

    """

    PaginatedResponseModel = _PaginatedResponse

    def __init__(
        self,
        *,
        token: str,
        base_url: str | httpx.URL,
        **kwargs: Any,
    ):
        super().__init__(
            token=token,
            base_url=httpx.URL(base_url).join(PRODUCTION_RUNS_API_SUB_PATH),
            **kwargs,
        )

    def _has_more(self, paginated_response: _PaginatedResponse) -> bool:
        return paginated_response.metadata.has_more

    def _next_page_call_args(
        self,
        url: str,
        params: dict[str, Any],
        paginated_response: _PaginatedResponse,
        **kwargs: Any,
    ) -> tuple[str, dict[str, Any], dict[str, Any]]:
        next_params = {**params, "cursor": paginated_response.metadata.next_cursor}
        return (url, next_params, kwargs)



from abc import abstractmethod
from datetime import datetime, timedelta
from typing import Any, Optional
from uuid import UUID

from pydantic import Field

import enlyze.models as user_models
from enlyze.api_clients.base import ApiBaseModel


class ProductionRunsApiModel(ApiBaseModel):
    """Base class for Production Runs API object models using pydantic

    All objects received from the Production Runs API are passed into models that derive
    from this class and thus use pydantic for schema definition and validation.

    """

    @abstractmethod
    def to_user_model(self, *args: Any, **kwargs: Any) -> Any:
        """Convert to a model that will be returned to the user."""


class OEEComponent(ProductionRunsApiModel):
    score: float
    time_loss: int

    def to_user_model(self) -> user_models.OEEComponent:
        """Convert into a :ref:`user model <user_models>`"""

        return user_models.OEEComponent(
            score=self.score,
            time_loss=timedelta(seconds=self.time_loss),
        )


class Product(ProductionRunsApiModel):
    code: str
    name: Optional[str]

    def to_user_model(self) -> user_models.Product:
        """Convert into a :ref:`user model <user_models>`"""

        return user_models.Product(
            code=self.code,
            name=self.name,
        )


class Quantity(ProductionRunsApiModel):
    unit: str | None
    value: float

    def to_user_model(self) -> user_models.Quantity:
        """Convert into a :ref:`user model <user_models>`"""

        return user_models.Quantity(
            unit=self.unit,
            value=self.value,
        )


class Machine(ApiBaseModel):
    name: str
    uuid: UUID


class ProductionRun(ProductionRunsApiModel):
    uuid: UUID
    machine: Machine = Field(alias="appliance")
    average_throughput: Optional[float]
    production_order: str
    product: Product
    start: datetime
    end: Optional[datetime]
    quantity_total: Optional[Quantity]
    quantity_scrap: Optional[Quantity]
    quantity_yield: Optional[Quantity]
    availability: Optional[OEEComponent]
    performance: Optional[OEEComponent]
    quality: Optional[OEEComponent]
    productivity: Optional[OEEComponent]

    def to_user_model(
        self, machines_by_uuid: dict[UUID, user_models.Machine]
    ) -> user_models.ProductionRun:
        """Convert into a :ref:`user model <user_models>`"""

        quantity_total = (
            self.quantity_total.to_user_model() if self.quantity_total else None
        )
        quantity_scrap = (
            self.quantity_scrap.to_user_model() if self.quantity_scrap else None
        )
        quantity_yield = (
            self.quantity_yield.to_user_model() if self.quantity_yield else None
        )
        availability = self.availability.to_user_model() if self.availability else None
        performance = self.performance.to_user_model() if self.performance else None
        quality = self.quality.to_user_model() if self.quality else None
        productivity = self.productivity.to_user_model() if self.productivity else None

        return user_models.ProductionRun(
            uuid=self.uuid,
            machine=machines_by_uuid[self.machine.uuid],
            average_throughput=self.average_throughput,
            production_order=self.production_order,
            product=self.product.to_user_model(),
            start=self.start,
            end=self.end,
            quantity_total=quantity_total,
            quantity_scrap=quantity_scrap,
            quantity_yield=quantity_yield,
            availability=availability,
            performance=performance,
            quality=quality,
            productivity=productivity,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from _typeshed import DataclassInstance
import typing
from dataclasses import is_dataclass
from types import UnionType

import pandas


def _flat_dataclass_schema(
    dataclass_obj_or_type: DataclassInstance | type[DataclassInstance],
    path_separator: str,
    _parent_path: list[str] = [],
) -> list[str]:
    """Derive flat schema of potentially nested dataclass ``dataclass_obj_or_type``"""

    flat: list[str] = []

    for field, typ in typing.get_type_hints(dataclass_obj_or_type).items():
        current_path = _parent_path + [field]
        field_types = (typ,)

        # expand union types (includes typing.Optional)
        origin_type = typing.get_origin(typ)
        if origin_type is UnionType or origin_type is typing.Union:
            field_types = typing.get_args(typ)

        for field_type in field_types:
            if is_dataclass(field_type):
                flat.extend(
                    _flat_dataclass_schema(field_type, path_separator, current_path)
                )
            elif field_type is not type(None):
                flat.append(path_separator.join(current_path))

    # dedupe while preserving order
    return list(dict.fromkeys(flat))


def dataframe_ensure_schema(
    df: pandas.DataFrame,
    dataclass_obj_or_type: DataclassInstance | type[DataclassInstance],
    path_separator: str = ".",
) -> pandas.DataFrame:
    """Add missing columns to ``df`` based on flattened dataclass schema"""

    flat_schema = _flat_dataclass_schema(
        dataclass_obj_or_type,
        path_separator=path_separator,
    )

    add_colums = set(flat_schema) - set(df.columns)
    remove_columns = set(df.columns) - set(flat_schema)

    return df.assign(
        **{col: None for col in add_colums},
    ).drop(
        columns=list(remove_columns),
    )



from collections import abc
from datetime import datetime
from functools import cache, reduce
from typing import Any, Iterator, Mapping, Optional, Sequence, Tuple, Union
from uuid import UUID

import enlyze.api_clients.timeseries.models as timeseries_api_models
import enlyze.models as user_models
from enlyze.api_clients.production_runs.client import ProductionRunsApiClient
from enlyze.api_clients.production_runs.models import ProductionRun
from enlyze.api_clients.timeseries.client import TimeseriesApiClient
from enlyze.constants import (
    ENLYZE_BASE_URL,
    MAXIMUM_NUMBER_OF_VARIABLES_PER_TIMESERIES_REQUEST,
    VARIABLE_UUID_AND_RESAMPLING_METHOD_SEPARATOR,
)
from enlyze.errors import EnlyzeError, ResamplingValidationError
from enlyze.iterable_tools import chunk
from enlyze.validators import (
    validate_datetime,
    validate_resampling_interval,
    validate_resampling_method_for_data_type,
    validate_start_and_end,
    validate_timeseries_arguments,
)

FETCHING_TIMESERIES_DATA_ERROR_MSG = "Error occurred when fetching timeseries data."


def _get_timeseries_data_from_pages(
    pages: Iterator[timeseries_api_models.TimeseriesData],
) -> Optional[timeseries_api_models.TimeseriesData]:
    try:
        timeseries_data = next(pages)
    except StopIteration:
        return None

    if not timeseries_data.columns:
        return None

    if "time" not in timeseries_data.columns:
        raise EnlyzeError("Timeseries API didn't return timestamps")

    for page in pages:
        timeseries_data.extend(page)

    return timeseries_data


def _get_variables_sequence_and_query_parameter_list(
    variables: Union[
        Sequence[user_models.Variable],
        Mapping[user_models.Variable, user_models.ResamplingMethod],
    ],
    resampling_interval: Optional[int],
) -> Tuple[Sequence[user_models.Variable], Sequence[str]]:
    if isinstance(variables, abc.Sequence) and resampling_interval is not None:
        raise ResamplingValidationError(
            "`variables` must be a mapping {variable: ResamplingMethod}"
        )

    if resampling_interval:
        validate_resampling_interval(resampling_interval)
        variables_sequence = []
        variables_query_parameter_list = []
        for variable, resampling_method in variables.items():  # type: ignore
            variables_sequence.append(variable)
            variables_query_parameter_list.append(
                f"{variable.uuid}"
                f"{VARIABLE_UUID_AND_RESAMPLING_METHOD_SEPARATOR}"
                f"{resampling_method.value}"
            )

            validate_resampling_method_for_data_type(
                resampling_method, variable.data_type
            )
        return variables_sequence, variables_query_parameter_list

    return variables, [str(v.uuid) for v in variables]  # type: ignore


class EnlyzeClient:
    """Main entrypoint for interacting with the ENLYZE platform.

    You should instantiate it only once and use it for all requests to make the best use
    of connection pooling. This client is thread-safe.

    :param token: API token for the ENLYZE platform

    """

    def __init__(self, token: str, *, _base_url: str | None = None) -> None:
        self._timeseries_api_client = TimeseriesApiClient(
            token=token,
            base_url=_base_url or ENLYZE_BASE_URL,
        )
        self._production_runs_api_client = ProductionRunsApiClient(
            token=token,
            base_url=_base_url or ENLYZE_BASE_URL,
        )

    def _get_sites(self) -> Iterator[timeseries_api_models.Site]:
        """Get all sites from the API"""
        return self._timeseries_api_client.get_paginated(
            "sites", timeseries_api_models.Site
        )

    @cache
    def get_sites(self) -> list[user_models.Site]:
        """Retrieve all :ref:`sites <site>` of your organization.

        :raises: |token-error|

        :raises: |generic-error|

        :returns: Sites of your organization
        :rtype: list[:class:`~enlyze.models.Site`]

        """
        return [site.to_user_model() for site in self._get_sites()]

    def _get_machines(self) -> Iterator[timeseries_api_models.Machine]:
        """Get all machines from the API"""
        return self._timeseries_api_client.get_paginated(
            "appliances", timeseries_api_models.Machine
        )

    @cache
    def get_machines(
        self, site: Optional[user_models.Site] = None
    ) -> list[user_models.Machine]:
        """Retrieve all :ref:`machines <machine>`, optionally filtered by site.

        :param site: Only get machines of this site. Gets all machines of the
            organization if None.

        :raises: |token-error|

        :raises: |generic-error|

        :returns: Machines
        :rtype: list[:class:`~enlyze.models.Machine`]

        """

        if site:
            sites_by_id = {site._id: site}
        else:
            sites_by_id = {site._id: site for site in self.get_sites()}

        machines = []
        for machine_api in self._get_machines():
            site_ = sites_by_id.get(machine_api.site)
            if not site_:
                continue

            machines.append(machine_api.to_user_model(site_))

        return machines

    def _get_variables(
        self, machine_uuid: UUID
    ) -> Iterator[timeseries_api_models.Variable]:
        """Get variables for a machine from the API."""
        return self._timeseries_api_client.get_paginated(
            "variables",
            timeseries_api_models.Variable,
            params={"appliance": str(machine_uuid)},
        )

    def get_variables(
        self, machine: user_models.Machine
    ) -> Sequence[user_models.Variable]:
        """Retrieve all variables of a :ref:`machine <machine>`.

        :param machine: The machine for which to get all variables.

        :raises: |token-error|

        :raises: |generic-error|

        :returns: Variables of ``machine``

        """
        return [
            variable.to_user_model(machine)
            for variable in self._get_variables(machine.uuid)
        ]

    def _get_paginated_timeseries(
        self,
        *,
        machine_uuid: str,
        start: datetime,
        end: datetime,
        variables: Sequence[str],
        resampling_interval: Optional[int],
    ) -> Iterator[timeseries_api_models.TimeseriesData]:
        params: dict[str, Any] = {
            "appliance": machine_uuid,
            "start_datetime": start.isoformat(),
            "end_datetime": end.isoformat(),
            "variables": ",".join(variables),
        }

        if resampling_interval:
            params["resampling_interval"] = resampling_interval

        return self._timeseries_api_client.get_paginated(
            "timeseries", timeseries_api_models.TimeseriesData, params=params
        )

    def _get_timeseries(
        self,
        start: datetime,
        end: datetime,
        variables: Union[
            Sequence[user_models.Variable],
            Mapping[user_models.Variable, user_models.ResamplingMethod],
        ],
        resampling_interval: Optional[int] = None,
    ) -> Optional[user_models.TimeseriesData]:
        variables_sequence, variables_query_parameter_list = (
            _get_variables_sequence_and_query_parameter_list(
                variables, resampling_interval
            )
        )

        start, end, machine_uuid = validate_timeseries_arguments(
            start, end, variables_sequence
        )

        try:
            chunks = chunk(
                variables_query_parameter_list,
                MAXIMUM_NUMBER_OF_VARIABLES_PER_TIMESERIES_REQUEST,
            )
        except ValueError as e:
            raise EnlyzeError(FETCHING_TIMESERIES_DATA_ERROR_MSG) from e

        chunks_pages = (
            self._get_paginated_timeseries(
                machine_uuid=machine_uuid,
                start=start,
                end=end,
                variables=chunk,
                resampling_interval=resampling_interval,
            )
            for chunk in chunks
        )

        timeseries_data_chunked = [
            _get_timeseries_data_from_pages(pages) for pages in chunks_pages
        ]

        if not timeseries_data_chunked or all(
            data is None for data in timeseries_data_chunked
        ):
            return None

        if any(data is None for data in timeseries_data_chunked) and any(
            data is not None for data in timeseries_data_chunked
        ):
            raise EnlyzeError(
                "The timeseries API didn't return data for some of the variables."
            )

        try:
            timeseries_data = reduce(lambda x, y: x.merge(y), timeseries_data_chunked)  # type: ignore # noqa
        except ValueError as e:
            raise EnlyzeError(FETCHING_TIMESERIES_DATA_ERROR_MSG) from e

        return timeseries_data.to_user_model(  # type: ignore
            start=start,
            end=end,
            variables=variables_sequence,
        )

    def get_timeseries(
        self,
        start: datetime,
        end: datetime,
        variables: Sequence[user_models.Variable],
    ) -> Optional[user_models.TimeseriesData]:
        """Get timeseries data of :ref:`variables <variable>` for a given time frame.

        Timeseries data for multiple variables can be requested at once. However, all
        variables must belong to the same machine.

        You should always pass :ref:`timezone-aware datetime
        <python:datetime-naive-aware>` objects to this method! If you don't, naive
        datetime objects will be assumed to be expressed in the local timezone of the
        system where the code is run.

        :param start: Start of the time frame for which to fetch timeseries data. Must
            not be before ``end``.
        :param end: End of the time frame for which to fetch timeseries data.
        :param variables: The variables for which to fetch timeseries data.

        :raises: |token-error|

        :raises: |generic-error|

        :returns: Timeseries data or ``None`` if the API returned no data for the
            request

        """

        return self._get_timeseries(start, end, variables)

    def get_timeseries_with_resampling(
        self,
        start: datetime,
        end: datetime,
        variables: Mapping[user_models.Variable, user_models.ResamplingMethod],
        resampling_interval: int,
    ) -> Optional[user_models.TimeseriesData]:
        """Get resampled timeseries data of :ref:`variables <variable>` for a given time frame.

        Timeseries data for multiple variables can be requested at once. However, all
        variables must belong to the same machine.

        You should always pass :ref:`timezone-aware datetime
        <python:datetime-naive-aware>` objects to this method! If you don't, naive
        datetime objects will be assumed to be expressed in the local timezone of the
        system where the code is run.

        :param start: Start of the time frame for which to fetch timeseries data. Must
            not be before ``end``.
        :param end: End of the time frame for which to fetch timeseries data.
        :param variables: The variables for which to fetch timeseries data along with a
            :class:`~enlyze.models.ResamplingMethod` for each variable. Resampling isn't
            supported for variables whose :class:`~enlyze.models.VariableDataType` is
            one of the `ARRAY` data types.
        :param resampling_interval: The interval in seconds to resample timeseries data
            with. Must be greater than or equal
            :const:`~enlyze.constants.MINIMUM_RESAMPLING_INTERVAL`.

        :raises: |token-error|

        :raises: |resampling-error|

        :raises: |generic-error|

        :returns: Timeseries data or ``None`` if the API returned no data for the
            request

        """  # noqa: E501
        return self._get_timeseries(start, end, variables, resampling_interval)

    def _get_production_runs(
        self,
        *,
        production_order: Optional[str] = None,
        product: Optional[str] = None,
        machine: Optional[UUID] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> Iterator[ProductionRun]:
        """Get production runs from the API."""

        filters = {
            "production_order": production_order,
            "product": product,
            "appliance": machine,
            "start": start.isoformat() if start else None,
            "end": end.isoformat() if end else None,
        }
        params = {k: v for k, v in filters.items() if v is not None}
        return self._production_runs_api_client.get_paginated(
            "production-runs", ProductionRun, params=params
        )

    def get_production_runs(
        self,
        *,
        production_order: Optional[str] = None,
        product: Optional[user_models.Product | str] = None,
        machine: Optional[user_models.Machine] = None,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> user_models.ProductionRuns:
        """Retrieve optionally filtered list of :ref:`production runs <production_run>`.

        :param machine: The machine for which to get all production runs.
        :param product: Filter production runs by product.
        :param production_order: Filter production runs by production order.

        :raises: |token-error|

        :raises: |generic-error|

        :returns: Production runs
        :rtype: :class:`~enlyze.models.ProductionRuns`

        """
        if start and end:
            start, end = validate_start_and_end(start, end)
        elif start:
            start = validate_datetime(start)
        elif end:
            end = validate_datetime(end)

        product_filter = (
            product.code if isinstance(product, user_models.Product) else product
        )
        machines_by_uuid = {a.uuid: a for a in self.get_machines()}
        return user_models.ProductionRuns(
            [
                production_run.to_user_model(machines_by_uuid)
                for production_run in self._get_production_runs(
                    machine=machine.uuid if machine else None,
                    production_order=production_order,
                    product=product_filter,
                    start=start,
                    end=end,
                )
            ]
        )



#: Base URL of the ENLYZE platform.
ENLYZE_BASE_URL = "https://app.enlyze.com"

#: URL sub-path where the Timeseries API is deployed on the ENLYZE platform.
TIMESERIES_API_SUB_PATH = "api/timeseries/v1/"

#: URL sub-path where the Production Runs API is deployed on the ENLYZE platform.
PRODUCTION_RUNS_API_SUB_PATH = "api/production-runs/v1/"

#: HTTP timeout for requests to the Timeseries API.
#:
#: Reference: https://www.python-httpx.org/advanced/timeouts/
HTTPX_TIMEOUT = 30.0

#: The separator to use when to separate the variable UUID and the resampling method
#: when querying timeseries data.
VARIABLE_UUID_AND_RESAMPLING_METHOD_SEPARATOR = "||"

#: The minimum allowed resampling interval when resampling timeseries data.
MINIMUM_RESAMPLING_INTERVAL = 10

#: The maximum number of variables that can be used in a single request when querying
#: timeseries data.
MAXIMUM_NUMBER_OF_VARIABLES_PER_TIMESERIES_REQUEST = 100

#: The user agent that the SDK identifies itself as when making HTTP requests
USER_AGENT = "enlyze-python"



from collections.abc import Generator

from httpx import Auth, Request, Response

from enlyze.errors import InvalidTokenError


class TokenAuth(Auth):
    """Token authentication scheme for use with ``httpx``

    :param token: API token for the ENLYZE platform

    """

    def __init__(self, token: str):
        if not isinstance(token, str):
            raise InvalidTokenError("Token must be a string")

        if not token:
            raise InvalidTokenError("Token must not be empty")

        self._auth_header = f"Token {token}"

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        """Inject token into authorization header"""
        request.headers["Authorization"] = self._auth_header
        yield request



from datetime import date, datetime
from typing import Any, Optional, Sequence
from uuid import UUID

import enlyze.models as user_models
from enlyze.api_clients.base import ApiBaseModel


class TimeseriesApiModel(ApiBaseModel):
    """Base class for Timeseries API object models using pydantic

    All objects received from the Timeseries API are passed into models that derive from
    this class and thus use pydantic for schema definition and validation.

    """

    pass


class Site(TimeseriesApiModel):
    id: int
    name: str
    address: str

    def to_user_model(self) -> user_models.Site:
        """Convert into a :ref:`user model <user_models>`"""

        return user_models.Site(
            _id=self.id,
            address=self.address,
            display_name=self.name,
        )


class Machine(TimeseriesApiModel):
    uuid: UUID
    name: str
    genesis_date: date
    site: int

    def to_user_model(self, site: user_models.Site) -> user_models.Machine:
        """Convert into a :ref:`user model <user_models>`"""

        return user_models.Machine(
            uuid=self.uuid,
            display_name=self.name,
            genesis_date=self.genesis_date,
            site=site,
        )


class Variable(TimeseriesApiModel):
    uuid: UUID
    display_name: Optional[str]
    unit: Optional[str]
    data_type: user_models.VariableDataType

    def to_user_model(self, machine: user_models.Machine) -> user_models.Variable:
        """Convert into a :ref:`user model <user_models>`."""

        return user_models.Variable(
            uuid=self.uuid,
            display_name=self.display_name,
            unit=self.unit,
            data_type=self.data_type,
            machine=machine,
        )


class TimeseriesData(TimeseriesApiModel):
    columns: list[str]
    records: list[Any]

    def extend(self, other: "TimeseriesData") -> None:
        """Add records from ``other`` after the existing records."""
        self.records.extend(other.records)

    def merge(self, other: "TimeseriesData") -> "TimeseriesData":
        """Merge records from ``other`` into the existing records."""
        slen, olen = len(self.records), len(other.records)
        if olen < slen:
            raise ValueError(
                "Cannot merge. Attempted to merge"
                f" an instance with {olen} records into an instance with {slen}"
                " records. The instance to merge must have a number"
                " of records greater than or equal to the number of records of"
                " the instance you're trying to merge into."
            )

        self.columns.extend(other.columns[1:])

        for s, o in zip(self.records, other.records[:slen]):
            if s[0] != o[0]:
                raise ValueError(
                    "Cannot merge. Attempted to merge records "
                    f"with mismatched timestamps {s[0]}, {o[0]}"
                )

            s.extend(o[1:])

        return self

    def to_user_model(
        self,
        start: datetime,
        end: datetime,
        variables: Sequence[user_models.Variable],
    ) -> user_models.TimeseriesData:
        return user_models.TimeseriesData(
            start=start,
            end=end,
            variables=variables,
            _columns=self.columns,
            _records=self.records,
        )



import json
from abc import ABC, abstractmethod
from collections.abc import Iterator
from functools import cache
from http import HTTPStatus
from typing import Any, Generic, TypeVar

import httpx
from pydantic import BaseModel, ValidationError

from enlyze._version import VERSION
from enlyze.auth import TokenAuth
from enlyze.constants import HTTPX_TIMEOUT, USER_AGENT
from enlyze.errors import EnlyzeError, InvalidTokenError

USER_AGENT_NAME_VERSION_SEPARATOR = "/"


@cache
def _construct_user_agent(
    *, user_agent: str = USER_AGENT, version: str = VERSION
) -> str:
    return f"{user_agent}{USER_AGENT_NAME_VERSION_SEPARATOR}{version}"


class ApiBaseModel(BaseModel):
    """Base class for ENLYZE platform API object models using pydantic

    All objects received from ENLYZE platform APIs are passed into models that derive
    from this class and thus use pydantic for schema definition and validation.

    """


class PaginatedResponseBaseModel(BaseModel):
    """Base class for paginated ENLYZE platform API responses using pydantic."""

    data: Any


#: TypeVar("M", bound=ApiBaseModel): Type variable serving as a parameter
# for API response model classes.
M = TypeVar("M", bound=ApiBaseModel)


#: TypeVar("R", bound=PaginatedResponseBaseModel) Type variable serving as a parameter
# for paginated response models.
R = TypeVar("R", bound=PaginatedResponseBaseModel)


class ApiBaseClient(ABC, Generic[R]):
    """Client base class encapsulating all interaction with all ENLYZE platform APIs.

    :param token: API token for the ENLYZE platform
    :param base_url: Base URL of the ENLYZE platform
    :param timeout: Global timeout for HTTP requests sent to the ENLYZE platform APIs

    """

    PaginatedResponseModel: type[R]

    def __init__(
        self,
        *,
        token: str,
        base_url: str | httpx.URL,
        timeout: float = HTTPX_TIMEOUT,
    ):
        self._client = httpx.Client(
            auth=TokenAuth(token),
            base_url=httpx.URL(base_url),
            timeout=timeout,
            headers={"user-agent": _construct_user_agent()},
        )

    @cache
    def _full_url(self, api_path: str) -> str:
        """Construct full URL from relative URL"""
        return str(self._client.build_request("", api_path).url)

    def get(self, api_path: str, **kwargs: Any) -> Any:
        """Wraps :meth:`httpx.Client.get` with defensive error handling

        :param api_path: Relative URL path inside the API name space (or a full URL)

        :raises: :exc:`~enlyze.errors.EnlyzeError` on request failure

        :raises: :exc:`~enlyze.errors.EnlyzeError` on non-2xx status code

        :raises: :exc:`~enlyze.errors.EnlyzeError` on non-JSON payload

        :returns: JSON payload of the response as Python object

        """

        try:
            response = self._client.get(api_path, **kwargs)
        except Exception as e:
            raise EnlyzeError(
                "Couldn't read from the ENLYZE platform API "
                f"(GET {self._full_url(api_path)})",
            ) from e

        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            if e.response.status_code in (
                HTTPStatus.UNAUTHORIZED,
                HTTPStatus.FORBIDDEN,
            ):
                raise InvalidTokenError
            else:
                raise EnlyzeError(
                    f"ENLYZE platform API returned error {response.status_code}"
                    f" (GET {self._full_url(api_path)})"
                ) from e

        try:
            return response.json()
        except json.JSONDecodeError as e:
            raise EnlyzeError(
                "ENLYZE platform API didn't return a valid JSON object "
                f"(GET {self._full_url(api_path)})",
            ) from e

    def _transform_paginated_response_data(self, data: Any) -> Any:
        """Transform paginated response data. Returns ``data`` by default.

        :param data: Response data from a paginated response

        :returns: An iterable of transformed data

        """
        return data

    @abstractmethod
    def _has_more(self, paginated_response: R) -> bool:
        """Indicates there is more data to fetch from the server.

        :param paginated_response: A paginated response model deriving from
            :class:`PaginatedResponseBaseModel`.

        """

    @abstractmethod
    def _next_page_call_args(
        self,
        *,
        url: str,
        params: dict[str, Any],
        paginated_response: R,
        **kwargs: Any,
    ) -> tuple[str, dict[str, Any], dict[str, Any]]:
        r"""Compute call arguments for the next page.

        :param url: The URL used to fetch the current page
        :param params: URL query parameters of the current page
        :param paginated_response: A paginated response model deriving from
            :class:`~enlyze.api_clients.base.PaginatedResponseBaseModel`
        :param \**kwargs: Keyword arguments passed into
            :py:meth:`~enlyze.api_clients.base.ApiBaseClient.get_paginated`

        :returns: A tuple of comprised of the URL, query parameters and keyword
            arguments to fetch the next page

        """

    def get_paginated(
        self, api_path: str, model: type[M], **kwargs: Any
    ) -> Iterator[M]:
        """Retrieve objects from a paginated ENLYZE platform API endpoint via HTTP GET.

        To add pagination capabilities to an API client deriving from this class, two
        abstract methods need to be implemented,
        :py:meth:`~enlyze.api_clients.base.ApiBaseClient._has_more` and
        :py:meth:`~enlyze.api_clients.base.ApiBaseClient._next_page_call_args`.
        Optionally, API clients may transform page data by overriding
        :py:meth:`~enlyze.api_clients.base.ApiBaseClient._transform_paginated_response_data`,
        which by default returns the unmodified page data.

        :param api_path: Relative URL path inside the API name space
        :param model: API response model class deriving from
            :class:`~enlyze.api_clients.base.ApiBaseModel`

        :raises: :exc:`~enlyze.errors.EnlyzeError` on invalid pagination schema

        :raises: :exc:`~enlyze.errors.EnlyzeError` on invalid data schema

        :raises: see :py:meth:`get` for more errors raised by this method

        :returns: Instances of ``model`` retrieved from the ``api_path`` endpoint

        """

        url = api_path
        params = kwargs.pop("params", {})

        while True:
            response_body = self.get(url, params=params, **kwargs)
            try:
                paginated_response = self.PaginatedResponseModel.model_validate(
                    response_body
                )
            except ValidationError as e:
                raise EnlyzeError(
                    f"Paginated response expected (GET {self._full_url(url)})"
                ) from e

            page_data = paginated_response.data
            if not page_data:
                break

            page_data = self._transform_paginated_response_data(page_data)

            for elem in page_data:
                try:
                    yield model.model_validate(elem)
                except ValidationError as e:
                    raise EnlyzeError(
                        f"ENLYZE platform API returned an unparsable {model.__name__} "
                        f"object (GET {self._full_url(api_path)})"
                    ) from e
            if not self._has_more(paginated_response):
                break

            url, params, kwargs = self._next_page_call_args(
                url=url,
                params=params,
                paginated_response=paginated_response,
                **kwargs,
            )



from .client import EnlyzeClient
from .errors import EnlyzeError, InvalidTokenError
from .models import Machine, ResamplingMethod, Site, Variable

__all__ = [
    "Machine",
    "EnlyzeClient",
    "EnlyzeError",
    "InvalidTokenError",
    "ResamplingMethod",
    "Site",
    "Variable",
]



class EnlyzeError(Exception):
    """Base class for all errors raised by the Timeseries SDK.

    In many situations this error will be raised with a concrete message describing the
    issue.

    """


class InvalidTokenError(EnlyzeError):
    """The ENLYZE platform token is not valid

    Possible reasons:

    - Invalid format (e.g. empty string)
    - Token has expired
    - Typo (e.g. check for excess whitespace)
    - Note: you can **not** use your password as a token

    """


class ResamplingValidationError(EnlyzeError):
    """Invalid request for resampling

    Timeseries data could not be requested from the ENLYZE platform. This may happen
    when a resampling method is incompatible with the variable data type or an invalid
    resampling interval is specified.

    """


class DuplicateDisplayNameError(EnlyzeError):
    """Variables with duplicate display names

    Resolving variable UUIDs to display names would result in ambiguity because
    multiple variables have the same display name. You should either fix the
    duplicate variable display names via the ENLYZE App or don't request them at
    the same time.
    """



from dataclasses import asdict, dataclass
from datetime import date, datetime, timedelta, timezone
from enum import Enum
from itertools import chain
from typing import Any, Iterator, Optional, Sequence
from uuid import UUID

import pandas

from enlyze.errors import DuplicateDisplayNameError
from enlyze.schema import dataframe_ensure_schema


@dataclass(frozen=True)
class Site:
    """Representation of a :ref:`site <site>` in the ENLYZE platform.

    Contains details about the site.

    """

    _id: int

    #: Display name of the site.
    display_name: str

    #: Postal address of the site. Doesn't follow a strict format.
    address: str


@dataclass(frozen=True)
class Machine:
    """Representation of a :ref:`machine <machine>` in the ENLYZE platform.

    Contains details about the machine.

    """

    #: Stable identifier of the machine.
    uuid: UUID

    #: Display name of the machine.
    display_name: str

    #: The date when the machine has been connected to the ENLYZE platform.
    genesis_date: date

    #: The site where the machine is located.
    site: Site


class VariableDataType(str, Enum):
    """Enumeration of variable data types. Compares to strings out-of-the-box:

    .. code-block:: pycon

        >>> VariableDataType.INTEGER == 'INTEGER'
        True

    """

    INTEGER = "INTEGER"
    FLOAT = "FLOAT"
    BOOLEAN = "BOOLEAN"
    STRING = "STRING"
    ARRAY_INTEGER = "ARRAY_INTEGER"
    ARRAY_FLOAT = "ARRAY_FLOAT"
    ARRAY_BOOLEAN = "ARRAY_BOOLEAN"
    ARRAY_STRING = "ARRAY_STRING"


class ResamplingMethod(str, Enum):
    """Resampling method to be used when resampling timeseries data."""

    FIRST = "first"
    LAST = "last"
    MAX = "max"
    MIN = "min"
    COUNT = "count"
    SUM = "sum"
    AVG = "avg"
    MEDIAN = "median"


@dataclass(frozen=True)
class Variable:
    """Representation of a :ref:`variable <variable>` in the ENLYZE platform.

    Contains details about the variable, but no timeseries data.

    """

    #: Stable identifier of the variable.
    uuid: UUID

    #: Display name of the variable.
    display_name: Optional[str]

    #: Unit of the measure that the variable represents.
    unit: Optional[str]

    #: The underlying data type of the variable.
    data_type: VariableDataType

    #: The machine on which this variable is read out.
    machine: Machine


@dataclass(frozen=True)
class TimeseriesData:
    """Result of a request for timeseries data."""

    #: Start of the requested time frame.
    start: datetime

    #: End of the requested time frame.
    end: datetime

    #: The variables for which timeseries data has been requested.
    variables: Sequence[Variable]

    _columns: list[str]
    _records: list[Any]

    def __len__(self) -> int:
        """Returns the number of resulting rows"""
        return len(self._records)

    def __repr__(self) -> str:
        """Summary of what was requested and how many records were retrieved"""
        return (
            "TimeseriesData: {:d} variables, {:d} records, start='{}' end='{}'".format(
                len(self.variables),
                len(self),
                self.start.isoformat(" "),
                self.end.isoformat(" "),
            )
        )

    def _display_names_as_column_names(self, columns: list[str]) -> list[str]:
        uuid_to_display_name = {
            str(var.uuid): var.display_name
            for var in self.variables
            if var.display_name
        }

        distinct_display_names = set(uuid_to_display_name.values())
        if len(uuid_to_display_name) != len(distinct_display_names):
            maybe_duplicate_display_names = list(uuid_to_display_name.values())
            for name in distinct_display_names:
                maybe_duplicate_display_names.remove(name)

            raise DuplicateDisplayNameError(
                ", ".join(
                    f"'{duplicate_display_name}'"
                    for duplicate_display_name in set(maybe_duplicate_display_names)
                )
            )

        return [uuid_to_display_name.get(var_uuid, var_uuid) for var_uuid in columns]

    def to_dicts(self, use_display_names: bool = False) -> Iterator[dict[str, Any]]:
        """Convert timeseries data into rows of :py:class:`dict`.

        Each row is returned as a dictionary with variable UUIDs as keys. Additionally,
        the ``time`` column is always present, containing :ref:`timezone-aware
        <python:datetime-naive-aware>` :py:class:`datetime.datetime` localized in UTC.

        :param use_display_names: Whether to return display names instead of variable
            UUIDs. If there is no display name, fall back to UUID.

        :raises: :exc:`~enlyze.errors.DuplicateDisplayNameError` when duplicate
            display names would be returned instead of UUIDs.

        :returns: Iterator over rows

        """
        time_column, *variable_columns = self._columns

        if use_display_names:
            variable_columns = self._display_names_as_column_names(variable_columns)

        for record in self._records:
            time_tuple = (
                time_column,
                datetime.fromisoformat(record[0]).astimezone(timezone.utc),
            )
            variable_tuples = zip(variable_columns, record[1:])
            yield dict(chain([time_tuple], variable_tuples))

    def to_dataframe(self, use_display_names: bool = False) -> pandas.DataFrame:
        """Convert timeseries data into :py:class:`pandas.DataFrame`

        The data frame will have an index named ``time`` that consists of
        :ref:`timezone-aware <python:datetime-naive-aware>`
        :py:class:`datetime.datetime` localized in UTC. Each requested variables is
        represented as a column named by its UUID.

        :param use_display_names: Whether to return display names instead of variable
            UUIDs. If there is no display name, fall back to UUID.

        :raises: :exc:`~enlyze.errors.DuplicateDisplayNameError` when duplicate
            display names would be returned instead of UUIDs.

        :returns: DataFrame with timeseries data indexed by time

        """

        time_column, *variable_columns = self._columns

        if use_display_names:
            variable_columns = self._display_names_as_column_names(variable_columns)

        df = pandas.DataFrame.from_records(
            self._records,
            columns=[time_column] + variable_columns,
            index="time",
        )
        df.index = pandas.to_datetime(df.index, utc=True, format="ISO8601")
        return df


@dataclass(frozen=True)
class OEEComponent:
    """Individual Overall Equipment Effectiveness (OEE) score

    This is calculated by the ENLYZE Platform based on a combination of real machine
    data and production order booking information provided by the customer.

    For more information, please check out https://www.oee.com

    """

    #: The score is expressed as a ratio between 0 and 1.0, with 1.0 meaning 100 %.
    score: float

    #: Unproductive time due to non-ideal production.
    time_loss: timedelta


@dataclass(frozen=True)
class Quantity:
    """Representation of a physical quantity"""

    #: Physical unit of quantity
    unit: str | None

    #: The quantity expressed in `unit`
    value: float


@dataclass(frozen=True)
class Product:
    """Representation of a product that is produced on a machine"""

    #: The identifier of the product
    code: str

    #: An optional human-friendly name of the product
    name: Optional[str] = None


@dataclass(frozen=True)
class ProductionRun:
    """Representation of a production run in the ENLYZE platform.

    Contains details about the production run.

    """

    #: The UUID of the production run
    uuid: UUID

    #: The machine the production run was executed on.
    machine: Machine

    #: The average throughput of the production run excluding downtimes.
    average_throughput: Optional[float]

    #: The identifier of the production order.
    production_order: str

    #: The identifier of the product that was produced.
    product: Product

    #: The begin of the production run.
    start: datetime

    #: The end of the production run.
    end: Optional[datetime]

    #: This is the sum of scrap and yield.
    quantity_total: Optional[Quantity]

    #: The amount of product produced that doesn't meet quality criteria.
    quantity_scrap: Optional[Quantity]

    #: The amount of product produced that can be sold.
    quantity_yield: Optional[Quantity]

    #: OEE component that reflects when the machine did not produce.
    availability: Optional[OEEComponent]

    #: OEE component that reflects how fast the machine has run.
    performance: Optional[OEEComponent]

    #: OEE component that reflects how much defects have been produced.
    quality: Optional[OEEComponent]

    #: Aggregate OEE score that comprises availability, performance and quality.
    productivity: Optional[OEEComponent]


class ProductionRuns(list[ProductionRun]):
    """Representation of multiple production runs."""

    def to_dataframe(self) -> pandas.DataFrame:
        """Convert production runs into :py:class:`pandas.DataFrame`

        Each row in the dataframe represents one production run. The ``start`` and
        ``end`` of every production run will be represented as :ref:`timezone-aware
        <python:datetime-naive-aware>` :py:class:`datetime.datetime` localized in UTC.

        :returns: DataFrame with production runs.
        """
        if not self:
            return pandas.DataFrame()

        path_separator = "."

        df = pandas.json_normalize([asdict(run) for run in self], sep=path_separator)
        df.start = pandas.to_datetime(df.start, utc=True, format="ISO8601")
        df.end = pandas.to_datetime(df.end, utc=True, format="ISO8601")

        return dataframe_ensure_schema(df, ProductionRun, path_separator=path_separator)



from typing import Iterable, Sequence, TypeVar

MINIMUM_CHUNK_SIZE = 1

T = TypeVar("T")


def chunk(seq: Sequence[T], chunk_size: int) -> Iterable[Sequence[T]]:
    if chunk_size < MINIMUM_CHUNK_SIZE:
        raise ValueError(f"{chunk_size=} is less than {MINIMUM_CHUNK_SIZE=}")

    return (seq[i : i + chunk_size] for i in range(0, len(seq), chunk_size))
